## Parameter file to generate SO-like data set

## Run ID - this will be used to create names
name_instrument = "SO"

#####################################################################
######################### Input sky #################################
#####################################################################
## can be two things:
##  1) Input fits file containing the sky maps to scan (maps will be loaded).
##  2) CAMB lensed cl file (.dat) containing lensed power spectra with
##      order ell, TT, EE, BB, TE (maps will be created on-the-fly).
## /!\ Option 2) is more for test purposes /!\
input_filename = "/global/cscratch1/sd/peloton/s4cmb/additional_files/SO_v2/IQU_noBB_freq150_large_v2.fits"

## Set do_pol to False if you want to load only intensity map.
do_pol = True

## If input_filename is a CAMB lensed cl file, the code generate maps at
## a resolution nside_in and convolved with
## a beam having this FWHM_in [arcmin].
## No effect if you provide fits file with maps instead of the CAMB file.
## /!\ this is more for test purposes /!\
fwhm_in = 0.0
nside_in = None
map_seed = None

# Remove either I, Q or U to remove possible leakages
no_ileak = False
no_quleak = False

# Set it to True if you are reading a map in Galactic coordinate.
# (Planck maps for example).
ext_map_gal = False

#####################################################################
######################### Instrument ################################
#####################################################################
## creates the data used to model the instrument (see s4cmb/instrument.py).
## Focal plane parameters: Number of crate plate,
## number of MUX board per crate,
## number of SQUID per MUX and number of pair of bolometers per SQUID.
ncrate = 4
ndfmux_per_crate = 7
nsquid_per_mux = 4
npair_per_squid = 28

## The size of the focal plane [cm].
fp_size = 60.

## Beam parameters of the bolometers.
## Full Width Half Maximum [arcmin].
# beams_large = {'27':9, '39':6.6, '90':2.8, '150':1.8, '220':1.2, '280':1}
# beams_small = {'27':91, '39':66, '90':28, '150':18, '220':12, '280':10}
fwhm = 1.8

## Seed used to generate angle of rotation of beam axes.
beam_seed = 58347

## Diameter of the focal plane [degree].
projected_fp_size = 3.

# Pointing model parameters of the telescope.
pm_name = "5params"

# Polarisation angle of the bolometers and HWP. Choose among CRHWP or stepped.
type_hwp = "CRHWP"

## HWP frequency [Hz] (only if CRHWP is used).
freq_hwp = 2.

## starting position of the HWP if type_hwp=CRHWP is used
## or the daily step change of the HWP if type_hwp=stepped is used.
## [deg].
angle_hwp = 0.

## Time-domain Noise level for the whole array for one CES.
## Default are noiseless simulations (i.e. = None).
## If you want to inject white noise on-the-fly, then specify a
## noise level in [u]K.sqrt(s). Careful the units has to be the same as
## the input map! Note also that it corresponds to the polarisation level.
## 8.1 / 152**.5 * 12**.5 i.e. noise level for 5 y of
## observations with 1/6 of efficiency. =
## 4.3 uk.arcmin @ nside=2048 @ 12 CES
## The multiplication by 12 comes from the fact that
## we group observation in 12 CES.
array_noise_level = 2.273

## Seed used to generate random numbers to simulate noise.
## From this single seed, we generate a list of seeds
## for all detectors. Has an effect only if array_noise_level is provided.
array_noise_seed = 487587

#####################################################################
######################### Scanning strategy #########################
#####################################################################
# A scanning strategy consists in defining the site of observation
# on earth for which we will make the observation, the region
# of the sky to observe, and the schedule of observations.
# (see s4cmb/scanning_strategy.py)

## Name of a pre-defined scanning strategy to define the boundaries
## of the scan: elevation, azimuth, and time. Only available for the:
## moment name_strategy = deep_patch.
name_strategy = "deep_patch"

## Number of CES to run. Should not exceed the number of CES contained in
## the defined scanning strategy set in name_strategy.
nces = 12

## Starting date for observations (The format is: YYYY/M/D HH:MM:SS)
start_date = "2013/1/1 00:00:00"

## Position of the telescope on Earth
telescope_longitude = "-67:46.816"
telescope_latitude = "-22:56.396"

## Elevation (Alt) of the telescope [meter]
telescope_elevation = 5200.

## Sampling frequency at which time samples are recorded [Hz]
sampling_freq = 15.

## Azimuth speed of the telescope [deg/s]
sky_speed = 0.4

## File containing time correction to UTC
ut1utc_fn = "/global/homes/p/peloton/s4cmb/s4cmb/data/ut1utc.ephem"

#####################################################################
######################### Output sky maps ###########################
#####################################################################
## Choose the projection of the output maps: healpix of flat
projection = "flat"

## If projection=healpix, choose the nside
nside_out = 2048

## If projection=flat, choose the pixel_size
pixel_size = 1.7

## Width of the output map in degree.
## If width < len(input scan), you will get a cropped map.
## if width > len(input scan), you will have your entire
## scan + plenty of zeros.
## A choice has to be made.
width = 60.

## Project time-domain data detector-by-detector to save a lot of memory.
## Use this only if you do not have detector-to-detector correlation!
mapping_perpair = True

#####################################################################
######################### Misc ######################################
#####################################################################
## language use to perform core computations.
language = "fortran"

## Verbose mode.
verbose = False



#!/usr/bin/python
"""
Script to simulate the hardware of a CMB experiment.
* focal plane
* pointing model parameters of the telescope
* beam parameters of the bolometers
* polarisation angle of the bolometers

Author: Julien Peloton, j.peloton@sussex.ac.uk
"""
from __future__ import division, absolute_import, print_function

import os
import copy
import glob
import datetime
import numpy as np

def coordinates_on_grid(pix_size=None, row_size=None,
                        nx=None, nx2=None,
                        max_points=None):
    """
    Return the x and y coordinates of points on a grid.
    The grid is centered on (0, 0).

    Parameters
    ----------
    pix_size : float, optional
        Size of each pixel. User should either provide
        `pix_size` or `row_size` (but not both at the same time).
    row_size : float, optional
        Size of each row. User should either provide
        `pix_size` or `row_size` (but not both at the same time).
    nx : int, optional
        Number of pixels per row/column. User should either provide
        `nx` or `nx2` (but not both at the same time).
    nx2 : int, optional
        Total number of pixels in the array. User should either provide
        `nx` or `nx2` (but not both at the same time).
    max_points : int, optional
        If nx2 is specified, `max_points` defines the maximum number of points
        to return. If None, set to `nx2`.

    Returns
    ----------
    coordinates : ndarray (2, nx[:max_points] * nx[:max_points])
        x and y coordinates of the pixels.

    Examples
    ----------
    Make a grid with 2 points per row, spaced by 1 unit
    >>> coordinates_on_grid(pix_size=1., nx=2)
    ... # doctest: +NORMALIZE_WHITESPACE
    array([[-0.5,  0.5, -0.5,  0.5],
           [-0.5, -0.5,  0.5,  0.5]])

    Same grid, but specifying the total number of points
    >>> coordinates_on_grid(pix_size=1., nx2=4)
    ... # doctest: +NORMALIZE_WHITESPACE
    array([[-0.5,  0.5, -0.5,  0.5],
           [-0.5, -0.5,  0.5,  0.5]])

    You can also specify the maximum number of points to return
    >>> coordinates_on_grid(pix_size=1., nx2=4, max_points=3)
    ... # doctest: +NORMALIZE_WHITESPACE
    array([[-0.5,  0.5, -0.5],
           [-0.5, -0.5,  0.5]])

    If you specify a total number of points which does not fit totally inside
    a squared grid, then it will return only the point you ask on a bigger grid
    Note that it works as if you would have specify max_points for a higher nx2
    >>> coordinates_on_grid(pix_size=1., nx2=3)
    ... # doctest: +NORMALIZE_WHITESPACE
    array([[-0.5,  0.5, -0.5],
           [-0.5, -0.5,  0.5]])

    You should specify either the number of pixel per row column (nx),
    or the total number of point in the grid but not both at the same time:
    >>> coordinates_on_grid(pix_size=1., nx=2, nx2=4)
    ... # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
    Traceback (most recent call last):
      ...
    AssertionError: You should specify either the number of pixel
    per row column (nx), or the total number of point in the grid (nx2).

    Idem for pix_size and row_size
    >>> coordinates_on_grid(pix_size=1., row_size=4., nx=2)
    ... # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
    Traceback (most recent call last):
      ...
    AssertionError: You should specify either the size of
    a pixel (pix_size), or the size of a row (row_size).
    """
    if (nx is None and nx2 is None) or (nx is not None and nx2 is not None):
        raise AssertionError('You should specify either the ' +
                             'number of pixel per row column (nx), or ' +
                             'the total number of point in the grid (nx2).\n')

    if (pix_size is None and row_size is None) or \
            (pix_size is not None and row_size is not None):
        raise AssertionError('You should specify either the ' +
                             'size of a pixel (pix_size), ' +
                             'or the size of a row (row_size).\n')

    if nx2 is not None:
        ## Look for the closest number with square root being an integer
        nx2_tmp = copy.copy(nx2)
        while True:
            nx = np.sqrt(nx2_tmp)
            if int(nx) == nx:
                nx = int(nx)
                break
            else:
                nx2_tmp += 1
    else:
        nx2 = nx**2

    if max_points is None:
        max_points = nx2

    if pix_size is None:
        pix_size = row_size / nx
    elif row_size is None:
        row_size = pix_size * nx

    ix1 = np.arange(nx)
    xs1 = (ix1 - (nx - 1.) / 2.) * pix_size
    x2, y2 = np.meshgrid(xs1, xs1)

    coordinates = np.array(
        (x2.flatten()[:max_points],
         y2.flatten()[:max_points]))

    return coordinates

def convert_pair_to_bolometer_position(xcoord_pairs, ycoord_pairs):
    """
    Return the position of bolometers given the position of pairs.

    Parameters
    ----------
    xcoord_pairs : ndarray
        Array of length `npair` containing the coordinate
        of the pairs of bolometers along the x axis.
    ycoord_pairs : ndarray
        Array of length `npair` containing the coordinate
        of the pairs of bolometers along the y axis.

    Returns
    ----------
    xcoord_bolometers : ndarray
        Array of length `2 * npair` containing the coordinate of
        the bolometers along the x axis.
    ycoord_bolometers : ndarray
        Array of length `2 * npair` containing the coordinate of
        the bolometers along the y axis.

    Examples
    ----------
    >>> fp = FocalPlane(verbose=False)
    >>> xp, yp = coordinates_on_grid(row_size=fp.fp_size, nx2=4)
    >>> print(xp, yp)
    [-15.  15. -15.  15.] [-15. -15.  15.  15.]
    >>> xb, yb = convert_pair_to_bolometer_position(xp, yp)
    >>> print(xb, yb) # doctest: +NORMALIZE_WHITESPACE
    [-15. -15.  15.  15. -15. -15.  15.  15.]
    [-15. -15. -15. -15.  15.  15.  15.  15.]
    """
    nbolometer = 2 * len(xcoord_pairs)
    xcoord_bolometers = np.dstack(
        (xcoord_pairs, xcoord_pairs)).reshape((1, nbolometer))[0]
    ycoord_bolometers = np.dstack(
        (ycoord_pairs, ycoord_pairs)).reshape((1, nbolometer))[0]

    return xcoord_bolometers, ycoord_bolometers

def show_focal_plane(bolo_xcoord, bolo_ycoord, bolo_polangle=None,
                     fn_out='plot_hardware_map_test.png',
                     save_on_disk=True, display=False):
    """
    Show the focal plane of the instrument, split in two panels:
    top and bottom bolometers

    Parameters
    ----------
    bolo_xcoord : 1d array
        Bolometers x coordinates in the focal plane.
    bolo_ycoord : 1d array
        Bolometers y coordinates in the focal plane.
    bolo_polangle : 1d array, optional
        Bolometer intrinsic polarisation angle orientation. If provided,
        it is used to color code the figure.
    fn_out : string, optional
        Name of the output file containing the plot of the focal plane.
        Provide the extension (format: png or pdf).
    save_on_disk : bool
        If True, save the plot on disk.
    display : bool
        If True, show the plot.

    Examples
    ---------
    >>> fp = FocalPlane(verbose=False)
    >>> show_focal_plane(fp.bolo_xcoord, fp.bolo_ycoord, fp.bolo_polangle,
    ...     save_on_disk=False, display=False)
    """
    if not display:
        import matplotlib as mpl
        mpl.use('Agg')
        import matplotlib.pyplot as pl
        pl.ioff()
    else:
        import matplotlib.pyplot as pl

    if bolo_polangle is None:
        bolo_polangle = np.ones_like(bolo_xcoord)

    fig, ax = pl.subplots(1, 2, figsize=(10, 5))
    ## Top pixel
    ax[0].scatter(bolo_xcoord[::2], bolo_ycoord[::2],
                  c=bolo_polangle[::2], alpha=1, s=30, cmap=pl.cm.jet)
    ax[0].scatter(bolo_xcoord[::2], bolo_ycoord[::2],
                  c='black', s=30, marker='|',
                  label='Top pixel', alpha=0.6)
    ax[0].set_ylabel('y position (cm)')
    ax[0].set_xlabel('x position (cm)')
    ax[0].set_title('Top pixels')

    ## Bottom pixel
    ax[1].scatter(bolo_xcoord[1::2], bolo_ycoord[1::2],
                  c=bolo_polangle[1::2], alpha=1, s=30, cmap=pl.cm.jet)
    ax[1].scatter(bolo_xcoord[1::2], bolo_ycoord[1::2],
                  c='black', s=30, marker='_',
                  label='Bottom pixel', alpha=0.6)
    ax[1].set_ylabel('y position (cm)')
    ax[1].set_xlabel('x position (cm)')
    ax[1].set_title('Bottom pixels')

    if save_on_disk:
        pl.savefig(fn_out)
        pl.clf()
    if display:
        pl.show()

def convert_cm_to_rad(xcm, ycm, conversion):
    """
    Convert positions in cm of the pairs of bolometers or bolometers
    in the focal plane into positions in radian.

    Parameters
    ----------
    xcm : ndarray
        x coordinates of pairs of bolometers or bolometers in cm.
    ycm : ndarray
        y coordinates of pairs of bolometers or bolometers in cm.
    conversion : float
        Conversion factor in rad/cm.

    Returns
    ----------
    xrad : ndarray
        x coordinates of pairs of bolometers or bolometers in radians.
        It has the same length as `xcm`.
    yrad : ndarray
        y coordinates of pairs of bolometers or bolometers in radians.
        It has the same length as `ycm`.

    Examples
    ----------
    Focal plane of 60 cm diameter and mirror
    giving a 3 deg projection on the sky by default
    >>> fp = FocalPlane(verbose=False)
    >>> projected_fp_size = 3. ## degrees
    >>> xcm, ycm = coordinates_on_grid(
    ...                 row_size=fp.fp_size, nx2=fp.npair)
    >>> print(convert_cm_to_rad(xcm, ycm,
    ...     conversion=projected_fp_size / fp.fp_size * np.pi / 180.))
    ...     # doctest: +NORMALIZE_WHITESPACE
    (array([-0.01308997,  0.01308997, -0.01308997,  0.01308997]),
     array([-0.01308997, -0.01308997,  0.01308997,  0.01308997]))

    """
    return np.array(xcm) * conversion, np.array(ycm) * conversion

def construct_beammap(beamprm, ct, cb, nx, pix_size):
    """
    Construct the pixel beam maps
    (sum and difference of bolometer beam maps)

    Parameters
    ----------
    beamprm : beam_model instance
        Instance of beam_model.
    ct : int
        Index of the top bolometer in the pair.
    cb : int
        Index of the bottom bolometer in the pair.
    nx : int
        Number of pixels per row/column (in pixel).
    pix_size : float
        Size of each pixel (in radian).

    Returns
    ----------
    summap : ndarray
        Beam map made of the sum of bolometer beam maps.
    diffmap : ndarray
        Beam map made of the difference of bolometer beam maps.

    Examples
    ----------
    Note that bolometers within the same pixel are neighbour bolometers
    that is (ct, cb) = (0, 1) for example.
    >>> fp = FocalPlane(verbose=False)
    >>> bm = BeamModel(fp, verbose=False)
    >>> pix_size = 0.5 / 180. * np.pi / 60. # 0.5 arcmin in rad
    >>> summap, diffmap = construct_beammap(
    ...     beamprm=bm, ct=0, cb=1, nx=4, pix_size=pix_size)
    >>> print(summap) # doctest: +NORMALIZE_WHITESPACE
    [[ 0.77520676  0.86809111  0.86809111  0.77520676]
     [ 0.86809111  0.97210474  0.97210474  0.86809111]
     [ 0.86809111  0.97210474  0.97210474  0.86809111]
     [ 0.77520676  0.86809111  0.86809111  0.77520676]]
    """
    # Translate beams to origin and maintain differential pointing
    dx = beamprm.xpos[ct] - beamprm.xpos[cb]
    dy = beamprm.ypos[ct] - beamprm.ypos[cb]

    tx = 0.5 * dx
    bx = -0.5 * dx
    ty = 0.5 * dy
    by = -0.5 * dy

    xy2f = coordinates_on_grid(pix_size=pix_size, nx=nx)

    tmap = gauss2d(xy2f, tx, ty,
                   beamprm.Amp[ct], beamprm.sig_1[ct],
                   beamprm.sig_2[ct],
                   beamprm.ellip_ang[ct]).reshape((nx, nx))

    bmap = gauss2d(xy2f, bx, by,
                   beamprm.Amp[cb], beamprm.sig_1[cb],
                   beamprm.sig_2[cb],
                   beamprm.ellip_ang[cb]).reshape((nx, nx))

    summap = 0.5 * (tmap + bmap)
    diffmap = 0.5 * (tmap - bmap)

    return summap, diffmap

def gauss2d(xy, x_0, y_0, Amp, sig_xp, sig_yp, psi):
    """
    2D Gaussian model for beams.

    Parameters
    ----------
    xy : 2d array
        Columns are projected coordinates (xproj,yproj)
    x_0: float
        x coordinate of the center of the Gaussian.
    y_0: float
        y coordinate of the center of the Gaussian.
    Amp: float
        Amplitude of the Gaussian.
    sig_xp: float
        Sigma for the Gaussian in x' coordinate system (rotated
        system if psi != 0).
    sig_yp: float
        Sigma for the Gaussian in y' coordinate system (rotated
        system if psi != 0).
    psi: float
        Angle between normal coordinate system and
        primed system (normal is primed if psi = 0).

    Returns
    ----------
    z : 1d array
        Flatten beam map of shape (xy[1].shape, )

    Examples
    ----------
    >>> fp = FocalPlane(verbose=False)
    >>> bm = BeamModel(fp, verbose=False)
    >>> pix_size = 0.5 / 180. * np.pi / 60. # 0.5 arcmin in rad
    >>> xy = coordinates_on_grid(pix_size=pix_size, nx=4)
    >>> gauss2d(xy, x_0=0, y_0=0, Amp=1.,
    ...     sig_xp=bm.sig_1[0],
    ...     sig_yp=bm.sig_2[0], psi=0)
    ... # doctest: +NORMALIZE_WHITESPACE
    array([ 0.77520676,  0.86809111,  0.86809111,  0.77520676,
            0.86809111,  0.97210474,  0.97210474,  0.86809111,
            0.86809111,  0.97210474,  0.97210474,  0.86809111,
            0.77520676,  0.86809111,  0.86809111,  0.77520676])
    """

    x_1 = xy[0, :] - x_0
    y_1 = xy[1, :] - y_0

    xy_1 = np.array([x_1, y_1])

    psi2 = -psi * np.pi / 180.0

    # x/y coordinates make an angle psi with the xp/yp input coordinates
    R = np.array(
        [[np.cos(psi2), -np.sin(psi2)], [np.sin(psi2), np.cos(psi2)]])
    p = np.dot(R, xy_1)

    u = p[0, :]**2 / (2 * sig_xp**2) + p[1, :]**2 / (2 * sig_yp**2)

    # Hide underflow by clipping beam function at -430dB level
    mask = u < 100
    z = Amp * np.exp(-u * mask) * mask

    return z

class Hardware():
    """ Class to load all the hardware and models of the instrument in once """
    def __init__(self,
                 ncrate=1, ndfmux_per_crate=1, nsquid_per_mux=1,
                 npair_per_squid=4, fp_size=60.,
                 fwhm=3.5, beam_seed=58347,
                 projected_fp_size=3.,
                 pm_name='5params',
                 type_hwp='CRHWP', freq_hwp=2., angle_hwp=0., verbose=False):
        """
        This class creates the data used to model the instrument:
        * focal plane
        * pointing model parameters of the telescope
        * beam parameters of the bolometers
        * polarisation angle of the bolometers

        Parameters
        ----------
        ncrate : int
            Number of crate plate.
        ndfmux_per_crate : int
            Number of MUX board per crate.
        nsquid_per_mux : int
            Number of SQUID per MUX.
        npair_per_squid : int
            Number of pair of bolometers per SQUID.
        fp_size : float, optional
            The size of the focal plane in cm. Default is 60 cm.
        fwhm : float, optional
            Full Width Half Maximum of the beam (in arcmin).
            Default = 3.5 arcmin.
        beam_seed : int
            Seed used to generate angle of rotation of beam axes.
            Default is 58347.
        projected_fp_size : float, optional
            Size of the focal plane on the sky (in degree). This has to
            do with the size of the mirror. Default = 3 degrees.
        pm_name : string, optional
            The pointing model to load. Currently, only the five-parameter
            pointing model (Mangum 2001) is implemented (pm_name = 5params).
        verbose : boolean, optional
            If True, print out a number of useful comments for verboseging.

        Examples
        ----------
        >>> instrument = Hardware()
        """
        self.focal_plane = FocalPlane(ncrate, ndfmux_per_crate,
                                      nsquid_per_mux, npair_per_squid,
                                      fp_size, verbose)

        self.beam_model = BeamModel(self.focal_plane, fwhm, beam_seed,
                                    projected_fp_size, verbose)

        self.pointing_model = PointingModel(pm_name)

        self.half_wave_plate = HalfWavePlate(type_hwp, freq_hwp, angle_hwp)

class FocalPlane():
    """ Class to handle the focal plane of the instrument. """
    def __init__(self,
                 ncrate=1, ndfmux_per_crate=1, nsquid_per_mux=1,
                 npair_per_squid=4, fp_size=60., verbose=False):
        """
        Initialise our focal plane.

        Note.
        The polarisation angle model consists in defining detector polarisation
        angle. The focal plane is cut in quadrants (Crate).
        Within a quadrant, pixels are categorized into two: Q and U pixels.
        Q and U pixels have 45 degrees difference in their polarisation angle,
        and form lines within quadrant. Each pixel contains a top and a
        bottom bolometer, with 90 degrees difference in the polarisation angle.
        Then, you go from one quadrant to another by a global 90 deg rotation
        of the polarisation angle starting with quadrant 0 having
        theta_{Q, top} = 0 deg.

        Parameters
        ----------
        ncrate : int
            Number of crate plate.
        ndfmux_per_crate : int
            Number of MUX board per crate.
        nsquid_per_mux : int
            Number of SQUID per MUX.
        npair_per_squid : int
            Number of pair of bolometers per SQUID.
        fp_size : float, optional
            The size of the focal plane in cm. Default is 60 cm.
        verbose : boolean, optional
            If True, print out a number of useful comments for verboseging.
        """
        self.ncrate = ncrate
        self.ndfmux_per_crate = ndfmux_per_crate
        self.nsquid_per_mux = nsquid_per_mux
        self.npair_per_squid = npair_per_squid

        ## Total number of pairs and bolometers in the focal plane
        self.npair = self.ncrate * self.ndfmux_per_crate * \
            self.nsquid_per_mux * self.npair_per_squid
        self.nbolometer = self.npair * 2

        self.fp_size = fp_size

        self.verbose = verbose

        self.make_focal_plane()

    def make_focal_plane(self):
        """
        Create the hardware map of the instrument,
        taht is the focal plane geometry,
        the bolometers id, the wiring, and so on.
        The terminology used here is taken from the Polarbear experiment.
        The hierarchy is the following:

        +-------------------------------------------------------------------+
        |CRATE -> DFMUX -> SQUID -> BOLOMETER (top & bottom)
        |  |        |        |          |
        |  v        v        v          v
        |  id       id       id         id, xCoordinate, yCoordinate,
        |                               IndexInFocalPlane, polangle_orientation
        |                               IndexInSquid.
        +--------------------------------------------------------------------+

        Examples
        ----------
        >>> fp = FocalPlane(verbose=True)
        Hardware map generated...
        """
        ## Retrieve coordinate of the pairs inside the focal plane
        # xcoord, ycoord = self.compute_pairs_coordinates(self.npair)
        xcoord, ycoord = coordinates_on_grid(row_size=self.fp_size,
                                             nx2=self.npair)

        ## Initialise
        self.crate_id, self.dfmux_id = [], []
        self.squid_id, self.bolo_id = [], []
        self.bolo_index_in_squid, self.bolo_index_in_fp = [], []
        self.bolo_xcoord, self.bolo_ycoord, self.bolo_polangle = [], [], []
        ## Construct the hardware map
        max_hit = False
        while max_hit is False:
            bolo_index = 0
            pair_index = 0
            squid_index = 0
            dfmux_index = 0
            for crate in range(self.ncrate):
                ## CRATE
                self.crate_id.append('Cr{:03}'.format(crate))

                for dfmux in range(self.ndfmux_per_crate):
                    ## DFMUX
                    self.dfmux_id.append('Cr{:03}Df{:03}'.format(
                            crate, dfmux_index))

                    for squid in range(self.nsquid_per_mux):
                        ## SQUID
                        self.squid_id.append('Cr{:03}Df{:03}Sq{:03}'.format(
                            crate, dfmux_index, squid_index))

                        for pair in range(self.npair_per_squid):
                            ## BOLOMETER

                            ## Split Q/U
                            boloQ = 2*pair
                            boloU = 2*pair + 1

                            if int(pair_index/np.sqrt(self.npair)) % 2 == 0:
                                shift = 0.
                                shiftinv = 45.
                            else:
                                shift = 45.
                                shiftinv = 0.

                            ## Top pixel
                            ## Position of the bolometer within the SQUID
                            self.bolo_index_in_squid.append(boloQ)
                            self.bolo_index_in_fp.append(bolo_index)
                            self.bolo_id.append(
                                'Cr{:03}Df{:03}Sq{:03}Bo{:03}t'.format(
                                    crate, dfmux_index, squid_index, boloQ))
                            self.bolo_xcoord.append(xcoord[pair_index])
                            self.bolo_ycoord.append(ycoord[pair_index])

                            ## Q/U pixels
                            # bolo in pairs are separated by 90 deg,
                            # and each quadran is separated by 90 deg.
                            if bolo_index % 4 == 0:
                                shift_ = 45.
                                shiftinv_ = 0.
                            else:
                                shift_ = 0.
                                shiftinv_ = 45.
                            if xcoord[pair_index] < 0 and \
                                    ycoord[pair_index] >= 0:
                                angle = shift
                            elif xcoord[pair_index] >= 0 and \
                                    ycoord[pair_index] >= 0:
                                angle = 90. + shift_
                            elif xcoord[pair_index] >= 0 and \
                                    ycoord[pair_index] <= 0:
                                angle = 180. + shiftinv
                            elif xcoord[pair_index] <= 0 and \
                                    ycoord[pair_index] <= 0:
                                angle = 270. + shiftinv_

                            self.bolo_polangle.append(angle)

                            ## Move in bolo space (2 bolos/pair)
                            bolo_index += 1

                            ## Bottom pixel
                            ## Top pixel
                            ## Position of the bolometer within the SQUID
                            self.bolo_index_in_squid.append(boloU)
                            self.bolo_index_in_fp.append(bolo_index)
                            self.bolo_id.append(
                                'Cr{:03}Df{:03}Sq{:03}Bo{:03}b'.format(
                                    crate, dfmux_index, squid_index, boloU))
                            self.bolo_xcoord.append(xcoord[pair_index])
                            self.bolo_ycoord.append(ycoord[pair_index])

                            ## 90 degree difference wrt top bolometer
                            self.bolo_polangle.append((angle + 90) % 360)

                            ## Move again in bolo space (2 bolos/pair)
                            bolo_index += 1

                            ## Move in pair space
                            pair_index += 1

                            ## Close the job if you hit the maximum number of
                            ## bolometers or pairs.
                            try:
                                assert bolo_index < self.nbolometer, \
                                    'Hardware map generated...'
                                assert pair_index < self.npair, \
                                    'Hardware map generated...'
                            except AssertionError as e:
                                if self.verbose:
                                    print(str(e))
                                max_hit = True
                                break
                        squid_index += 1
                    dfmux_index += 1

    def get_indices(self, name='Cr'):
        """
        Returns Cr(ate), Df(mux) or Sq(uid) indices.

        Parameters
        ----------
        name : string
            Which component you want the indices.
            Should be in ['Cr', 'Sq', 'Df'].
            Note that if you want bolometer indices, just use
            self.bolo_index_in_fp (for global indices in the focal plane) or
            self.bolo_index_in_squid (for local indices in SQUIDs).

        Returns
        ----------
        indices : list of int
            List containing the indices of the component for all bolometers.

        Examples
        ----------
        >>> fp = FocalPlane()
        >>> squid_indices = fp.get_indices('Sq')
        >>> print(squid_indices)
        [0, 0, 0, 0, 0, 0, 0, 0]

        """
        assert name in ['Cr', 'Sq', 'Df'], \
            ValueError("name must be in ['Cr', 'Sq', 'Df'].")

        indices = [
            int(comp.split(name)[-1][:3]) for comp in self.bolo_id]

        return indices

class BeamModel():
    """ Class to handle the beams of the detectors """
    def __init__(self,
                 focal_plane, fwhm=3.5, beam_seed=58347,
                 projected_fp_size=3., verbose=False):
        """
        Parameters
        ----------
        focal_plane : focal_plane instance
            Instance of focal_plane containing focal plane parameters.
        fwhm : float, optional
            Full Width Half Maximum of the beam (in arcmin).
            Default = 3.5 arcmin.
        beam_seed : int
            Seed used to generate angle of rotation of beam axes.
            Default is 58347.
        projected_fp_size : float, optional
            Size of the focal plane on the sky (in degree). This has to
            do with the size of the mirror. Default = 3 degrees.
        verbose : boolean, optional
            If True, print out a number of useful comments for verboseging.
        """
        ## Focal plane parameters
        self.focal_plane = focal_plane

        ## Beam model and mirror parameters
        self.fwhm = fwhm
        self.beam_seed = beam_seed
        self.projected_fp_size = projected_fp_size

        ## Paths and names
        self.verbose = verbose

        self.beamprm = self.generate_beam_parameters()

    def generate_beam_parameters(self):
        """
        Construct the beam parameters such as position of the centroids,
        ellipticity, angle of rotation, and associated errors.

        Returns
        ----------
        beamprm_fields : dictionary
            Dictionary containing beam parameters

        Examples
        ----------
        >>> fp = FocalPlane(verbose=False)
        >>> bm = BeamModel(fp, verbose=False)
        >>> bm.generate_beam_parameters()
        >>> print(bm.xpos) # doctest: +NORMALIZE_WHITESPACE
        [-0.01308997 -0.01308997  0.01308997  0.01308997
         -0.01308997 -0.01308997  0.01308997  0.01308997]
        """

        beamprm_header = ['Amp', 'Amp_err',
                          'ellip_ang', 'ellip_ang_err',
                          'sig_1', 'sig_1_err',
                          'sig_2', 'sig_2_err',
                          'xpos', 'xpos_err',
                          'ypos', 'ypos_err']

        for b in beamprm_header:
            setattr(self, b, np.zeros(self.focal_plane.nbolometer))

        ## Position of the bolometers
        ## (bolometers cm -> bolometers radians)
        self.xpos, self.ypos = convert_cm_to_rad(
            self.focal_plane.bolo_xcoord, self.focal_plane.bolo_ycoord,
            conversion=self.projected_fp_size /
            self.focal_plane.fp_size * np.pi / 180.)

        ## Generate Gaussian beams.
        ## FWHM arcmin -> FWHM rad -> sigma rad
        FWHM_rad = self.fwhm / 60. * np.pi / 180.
        sigma_rad = FWHM_rad / np.sqrt(8 * np.log(2))

        self.sig_1 = np.ones(self.focal_plane.nbolometer) * sigma_rad
        self.sig_2 = np.ones(self.focal_plane.nbolometer) * sigma_rad

        ## Angle of rotation for the ellipses.
        ## Between -90 and 90 degrees.
        state = np.random.RandomState(self.beam_seed)
        self.ellip_ang = state.uniform(-90, 90, self.focal_plane.nbolometer)

        ## Amplitude of the beams
        ## Default is one.
        self.Amp = np.ones(self.focal_plane.nbolometer)

class PointingModel():
    """ Class to handle the pointing model of the telescope """
    def __init__(self, pm_name='5params'):
        """
        We focus on a five-parameter pointing model (Mangum 2001) to
        characterize the relationship between the telescope's encoder
        readings and its true boresight pointing on the sky.
        The parameters described in this reference are

        * IA, the azimuth encoder zero offset,
        * IE, the elevation encoder zero offset,
        * CA, the collimation error of the electromagnetic axis,
        * AN, the azimuth axis offset/misalignment (north-south) and
        * AW, the azimuth offset/misalignment (east-west).

        Parameters
        ----------
        pm_name : string, optional
            The pointing model to load. Currently, only the five-parameter
            pointing model (Mangum 2001) is implemented (pm_name = 5params).

        Examples
        --------
        >>> pm = PointingModel()
        >>> [(i, str(round(j,3))) for i, j in zip(
        ...     pm.allowed_params.split(), pm.value_params)]
        ... # doctest: +NORMALIZE_WHITESPACE
        [('ia', '-10.285'), ('ie', '8.74'),
         ('ca', '-15.598'), ('an', '-0.51'),
         ('aw', '0.109')]

        >>> pm = PointingModel(pm_name='super-model-trop-bien')
        ... # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        Traceback (most recent call last):
         ...
        ValueError: Only the five-parameter pointing model
        (Mangum 2001) is implemented for the moment (pm_name = 5params)
        """
        self.pm_name = pm_name

        if self.pm_name == '5params':
            self.five_parameter_pointing_model()
        else:
            raise ValueError('Only the five-parameter ' +
                             'pointing model (Mangum 2001) is implemented ' +
                             'for the moment (pm_name = 5params)')

    def five_parameter_pointing_model(self):
        """
        Parameters based on Polarbear configuration.
        """
        self.allowed_params = 'ia ie ca an aw'

        self.value_params = np.array([-10.28473073, 8.73953334,
                                      -15.59771781, -0.50977716, 0.10858016])

        ## Set this to zero for the moment
        self.RMS_AZ = 0.0
        self.RMS_EL = 0.0
        self.RMS_RESID = 0.0

class HalfWavePlate():
    """ Class to handle the Half-Wave Plate (HWP) """
    def __init__(self, type_hwp='CRHWP', freq_hwp=2., angle_hwp=0.):
        """
        This class provides routines to compute the HWP angles.
        This can be use later to generate timestreams.

        Parameters
        ----------
        type_hwp : string, optional
            The type of HWP that you want to mount on your instrument.
            * CRWHP: continously rotating HWP.
            * stepped: stepped HWP (once a CES).
        freq_hwp : float, optional
            The frequency of rotation of the HWP in Hz.
        angle_hwp : float, optional
            The offset of the HWP in degree.

        """
        self.type_hwp = type_hwp
        self.freq_hwp = freq_hwp
        self.angle_hwp = angle_hwp

        if self.type_hwp not in ['CRHWP', 'stepped']:
            raise ValueError("`type_hwp` has to be 'CRHWP' or 'stepped'.")

        if self.type_hwp is 'stepped' and freq_hwp != 0.0:
            raise AssertionError("You cannot have a stepped HWP and non-" +
                                 "zero frequency! set freq_hwp=0.0 " +
                                 "if you want a stepped HWP.")

    def compute_HWP_angles(self, sample_rate=1., size=1):
        """
        Generate HWP angles which can be use later to generate timestreams.

        Parameters
        ----------
        sample_rate : float, optional
            Sample rate of the detectors
        size : int, optional
            Length of the vector of angles (number of time samples).

        Returns
        ----------
        HWP_angles : ndarray
            1d array of size `size` containting the HWP angles in radian.

        Examples
        Continously rotating HWP at 2 Hz starting at 0 degree.
        >>> hwp = HalfWavePlate(type_hwp='CRHWP', freq_hwp=2., angle_hwp=0.)
        >>> hwp.compute_HWP_angles(sample_rate=100., size=10)
        ... # doctest: +NORMALIZE_WHITESPACE
        array([ 0.        ,  0.12566371,  0.25132741,  0.37699112,  0.50265482,
            0.62831853,  0.75398224,  0.87964594,  1.00530965,  1.13097336])

        Stepped HWP with 30 degrees angle
        >>> hwp = HalfWavePlate(type_hwp='stepped',
        ...     freq_hwp=0.0, angle_hwp=30.)
        >>> hwp.compute_HWP_angles(sample_rate=100., size=10)
        ... # doctest: +NORMALIZE_WHITESPACE
        array([ 0.52359878,  0.52359878,  0.52359878,  0.52359878,  0.52359878,
            0.52359878,  0.52359878,  0.52359878,  0.52359878,  0.52359878])

        For a stepped HWP, the frequency must be zero
        >>> hwp = HalfWavePlate(type_hwp='stepped',
        ...     freq_hwp=1.0, angle_hwp=30.)
        ... # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        Traceback (most recent call last):
            ...
        AssertionError: You cannot have a stepped HWP and non-zero frequency!
        set freq_hwp=0.0 if you want a stepped HWP.
        """
        angle = self.angle_hwp * np.pi / 180.

        HWP_angles = np.array(
            [angle + t * (self.freq_hwp / sample_rate) *
             2. * np.pi for t in range(size)])

        return HWP_angles

    def update_hardware(self, new_type_hwp, new_freq_hwp, new_angle_hwp):
        """
        Change the behaviour of the HWP.

        Parameters
        ----------
        new_type_hwp : string, optional
            The type of HWP that you want to mount on your instrument.
            * CRWHP: continously rotating HWP.
            * stepped: stepped HWP (once a CES).
        new_freq_hwp : float, optional
            The frequency of rotation of the HWP in Hz.
        new_angle_hwp : float, optional
            The offset of the HWP in degree.

        Examples
        ----------
        Continously rotating HWP at 2 Hz starting at 0 degree.
        >>> hwp = HalfWavePlate(type_hwp='CRHWP', freq_hwp=2., angle_hwp=0.)
        >>> hwp.compute_HWP_angles(sample_rate=100., size=10)
        ... # doctest: +NORMALIZE_WHITESPACE
        array([ 0.        ,  0.12566371,  0.25132741,  0.37699112,  0.50265482,
            0.62831853,  0.75398224,  0.87964594,  1.00530965,  1.13097336])

        For some reason, our HWP died!
        >>> hwp.update_hardware(new_type_hwp='stepped',
        ...     new_freq_hwp=0.0, new_angle_hwp=0.0)
        >>> hwp.compute_HWP_angles(sample_rate=100., size=10)
        ... # doctest: +NORMALIZE_WHITESPACE
        array([ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.])

        """
        self.type_hwp = new_freq_hwp
        self.freq_hwp = new_freq_hwp
        self.angle_hwp = new_angle_hwp


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from s4cmb import instrument
from s4cmb import scanning_strategy
from s4cmb import input_sky
from s4cmb import detector_pointing
from s4cmb import tod
from s4cmb import systematics
from s4cmb import config_s4cmb
from s4cmb import xpure



#!/usr/bin/python
"""
Module to simulate pointing of the focal plane detectors.

Reminder mostly for myself:
Az/El specify the direction of a point on the celestial sphere in the
horizontal coordinate system (a spherical coordinate system).
RA/Dec specify the direction of a point on the celestial sphere in the
equatorial coordinate system.

Author: Julien Peloton, j.peloton@sussex.ac.uk
"""
from __future__ import division, absolute_import, print_function

import healpy as hp
import numpy as np
from numpy import cos
from numpy import sin
from numpy import tan
from pyslalib import slalib

from s4cmb.detector_pointing_f import detector_pointing_f

sec2deg = 360.0/86400.0
d2r = np.pi / 180.0
ASTROMETRIC_GEOCENTRIC = 0
APPARENT_GEOCENTRIC = 1
APPARENT_TOPOCENTRIC = 2

def get_ut1utc(ut1utc_fn, mjd):
    """ Return the time correction to UTC.

    Those are computed from 01 01 2012 (MJD=55927) to 01 01 2018 (MJD=58119).
    Unfortunately the database used to compute those is currently down
    (http://maia.usno.navy.mil/cgi-bin/search.cgi), so for date > 01 01 2018,
    the correction applied will be the one on 01 01 2018.

    Parameters
    ----------
    ut1utc_fn : string
        Filename where ut1utc are stored.
    mjd : float
        Date (in MJD) to correct for.

    Returns
    ----------
    ut1utc : float
        Contain the time correction to apply to MJD values.

    Examples
    ----------
    >>> fn = 's4cmb/data/ut1utc.ephem'
    >>> print(round(get_ut1utc(fn, mjd=56293), 3))
    0.277

    For date > 01 01 2018 (MJD=58119), we do not have data currently.
    So by default it will return the correction for 01 01 2018.
    >>> print(round(get_ut1utc(fn, mjd=58119), 3))
    0.128
    >>> print(round(get_ut1utc(fn, mjd=59119), 3))
    0.128
    """
    if mjd <= 58119:
        umjds, ut1utcs = np.loadtxt(ut1utc_fn, usecols=(1, 2)).T
        uindex = np.searchsorted(umjds, mjd)
        ut1utc = ut1utcs[uindex]
    else:
        mjd_tmp = 58119
        umjds, ut1utcs = np.loadtxt(ut1utc_fn, usecols=(1, 2)).T
        uindex = np.searchsorted(umjds, mjd_tmp)
        ut1utc = ut1utcs[uindex]

    return ut1utc

class Pointing():
    """ Class to handle detector pointing """
    def __init__(self, az_enc, el_enc, time, value_params,
                 allowed_params='ia ie ca an aw',
                 ra_src=0.0, dec_src=0.0, lat=-22.958,
                 ut1utc_fn='s4cmb/data/ut1utc.ephem'):
        """
        Apply pointing model with parameters `value_params` and
        names `allowed_params` to encoder az,el. Order of terms is
        `value_params` is same as order of terms in `allowed_params`.

        Full list of parameters (Thx Fred!):
            an:  azimuth axis tilt north of vertical
            aw:  azimuth axis tilt west of vertical
            an2:  potato chip
            aw2:  potato chip
            npae:  not parallel azimuth/elevation
            ca:  angle of beam to boresight in azimuth
            ia:  azimuth encoder zero
            ie:  elevation encoder zero + angle of beam
                to boresight in elevation
            tf:  cosine flexure
            tfs:  sine flexure
            ref:  refraction
            dt:  timing error in seconds (requires lat argument)
            elt:  time from start elevation correction
            ta1: linear order structural thermal warping in azimuth
            te1: linear order structural thermal warping in elevation
            sa,sa2: solar radiation structural warping in azimuth
            se,se2: solar radiation structural warping in elevation

        Parameters
        ----------
        az_enc : 1d array
            Encoder azimuth in radians.
        el_enc : 1d array
            Encoder elevation in radians.
        time : 1d array
            Encoder time (UTC) in mjd
        value_params : 1d array
            Value of the pointing model parameters (see instrument.py).
            In degrees (see below for full description)
        allowed_params : list of string, optional
            Name of the pointing model parameters used in `value_params`.
        ra_src : float, optional
            RA of the reference (keep it to zero).
        dec_src : float, optional
            Dec of the reference (keep it to zero).
        lat : float, optional
            Latitude of the telescope, in degree.
        ut1utc_fn : string, optional
            File containing time correction to UTC.
            The \Delta{UT} (UT1-UTC) is tabulated in IERS circulars
            and elsewhere. It increases by exactly one second at the end
            of each UTC leap second, introduced in order to
            keep \Delta{UT} within \pm 0s.9.
            The 'sidereal \Delta{UT}' which forms part of AOPRMS(13)
            is the same quantity, but converted from solar to sidereal
            seconds and expressed in radians.
            WTF?

        Examples
        ----------
        See hardware.py for more information on the pointing model.
        >>> allowed_params, value_params, az_enc, el_enc, time = \
            load_fake_pointing()
        >>> pointing = Pointing(az_enc, el_enc, time, value_params,
        ...     allowed_params, lat=-22.)
        >>> print(round(az_enc[2], 2), round(pointing.az[2], 2))
        0.13 0.12
        """
        self.az_enc = az_enc
        self.el_enc = el_enc
        self.time = time
        self.value_params = value_params
        self.allowed_params = allowed_params
        self.lat = lat * d2r
        self.ut1utc_fn = ut1utc_fn
        self.ra_src = ra_src
        self.dec_src = dec_src

        self.ut1utc = get_ut1utc(self.ut1utc_fn, self.time[0])

        ## Initialise the object
        self.az, self.el = self.apply_pointing_model()
        self.azel2radec()

    def apply_pointing_model(self):
        """
        Apply pointing corrections specified by the pointing model.

        Returns
        ----------
        az : 1d array
            The corrected azimuth in arcminutes.
        el : 1d array
            The corrected elevation in arcminutes.
        """
        assert len(self.value_params) == len(self.allowed_params.split()), \
            AssertionError("Vector containing parameters " +
                           "(value_params) has to have the same " +
                           "length than the vector containing names " +
                           "(allowed_params).")

        ## Here are many parameters defining a pointing model.
        ## Of course, we do not use all of them. They are zero by default,
        ## and only those specified by the user will be used.
        params = {p: 0.0 for p in ['an', 'aw', 'an2', 'aw2', 'an4',
                                   'aw4', 'npae', 'ca', 'ia', 'ie', 'tf',
                                   'tfs', 'ref', 'dt', 'elt', 'ta1',
                                   'te1', 'sa', 'se', 'sa2',
                                   'se2', 'sta', 'ste', 'sta2', 'ste2']}

        for param in params:
            if param in self.allowed_params.split():
                index = self.allowed_params.split().index(param)
                params[param] = self.value_params[index]

        params['dt'] *= sec2deg

        ## Azimuth
        azd = -params['an'] * sin(self.az_enc) * sin(self.el_enc)
        azd -= params['aw'] * cos(self.az_enc) * sin(self.el_enc)

        azd -= -params['an2'] * sin(2 * self.az_enc) * sin(self.el_enc)
        azd -= params['aw2'] * cos(2 * self.az_enc) * sin(self.el_enc)

        azd -= -params['an4'] * sin(4 * self.az_enc) * sin(self.el_enc)
        azd -= params['aw4'] * cos(4 * self.az_enc) * sin(self.el_enc)

        azd += params['npae'] * sin(self.el_enc)
        azd -= params['ca']
        azd += params['ia'] * cos(self.el_enc)

        azd += params['dt'] * (
            -sin(self.lat) + cos(self.az_enc) *
            cos(self.lat) * tan(self.el_enc))

        ## Elevation
        eld = params['an'] * cos(self.az_enc)
        eld -= params['aw'] * sin(self.az_enc)
        eld -= params['an2'] * cos(2 * self.az_enc)
        eld -= params['aw2'] * sin(2 * self.az_enc)
        eld -= params['an4'] * cos(4 * self.az_enc)
        eld -= params['aw4'] * sin(4 * self.az_enc)

        eld -= params['ie']
        eld += params['tf'] * cos(self.el_enc)
        eld += params['tfs'] * sin(self.el_enc)
        eld -= params['ref'] / tan(self.el_enc)

        eld += -params['dt'] * cos(self.lat) * sin(self.az_enc)

        eld += params['elt'] * (self.time - np.min(self.time))

        ## Convert back in radian and apply to the encoder values.
        azd *= np.pi / (180.0 * 60.)
        eld *= np.pi / (180.0 * 60.)

        azd /= np.cos(self.el_enc)

        az = self.az_enc - azd
        el = self.el_enc - eld

        return az, el

    def azel2radec(self):
        """
        Given Az/El, time, and time correction returns RA/Dec, parallactic
        angles, and quaternions (used later to rotate detector coordinates).

        Examples
        ----------
        Go from az/el -> ra/dec
        >>> allowed_params, value_params, az_enc, el_enc, time = \
            load_fake_pointing()
        >>> pointing = Pointing(az_enc, el_enc, time, value_params,
        ...     allowed_params, lat=-22.)
        >>> print(round(pointing.ra[2], 2), round(pointing.dec[2], 2))
        0.7 0.66
        """
        self.ra, self.dec, self.pa = self.azel2radecpa()
        v_ra = self.ra
        v_dec = self.dec
        v_pa = self.pa
        v_ra_src = self.ra_src
        v_dec_src = self.dec_src

        self.meanpa = np.median(v_pa)

        self.quaternion = Quaternion(v_ra, v_dec, v_pa,
                                     v_ra_src, v_dec_src)

        q = self.quaternion.offset_radecpa_makequat()

        assert q.shape == (self.az.size, 4), \
            AssertionError("Wrong size for the quaternions!")

        self.q = q

    def azel2radecpa(self):
        """
        Given Az/El, time, and time correction returns RA/Dec and parallactic
        angles.

        Examples
        ----------
        Go from az/el -> ra/dec/pa
        >>> allowed_params, value_params, az_enc, el_enc, time = \
            load_fake_pointing()
        >>> pointing = Pointing(az_enc, el_enc, time, value_params,
        ...     allowed_params, lat=-22.)
        >>> ra, dec, pa = pointing.azel2radecpa()
        >>> print(round(ra[0], 2), round(dec[0], 2), round(pa[0], 2))
        0.56 0.67 3.13
        """
        ## TODO pass lon, lat, etc from the ScanningStrategy module!
        converter = Azel2Radec(self.time[0], self.ut1utc)
        vconv = np.vectorize(converter.azel2radecpa)
        ra, dec, pa = vconv(self.time, self.az, self.el)
        return ra, dec, pa

    def radec2azel(self):
        """
        Given RA/Dec, time, and time correction returns Az/El.

        Examples
        ----------
        Make sure that we go back to our feet az/el -> ra/dec -> az/el
        >>> allowed_params, value_params, az_enc, el_enc, time = \
            load_fake_pointing()
        >>> pointing = Pointing(az_enc, el_enc, time, value_params,
        ...     allowed_params, lat=-22.)
        >>> az, el = pointing.radec2azel()
        >>> assert np.all(np.round(az[2:4],2) == np.round(pointing.az[2:4],2))
        >>> assert np.all(np.round(el[2:4],2) == np.round(pointing.el[2:4],2))
        """
        ## TODO pass lon, lat, etc from the ScanningStrategy module!
        converter = Azel2Radec(self.time[0], self.ut1utc)
        vconv = np.vectorize(converter.radec2azel)
        az, el = vconv(self.time, self.ra, self.dec)
        return az, el

    def offset_detector(self, azd, eld):
        """
        To compute RA/Dec of each detector from az/el, it is much
        faster to use the quaternions. This routine does it for you.

        Parameters
        ----------
        azd : 1d array
            The azimuth array for the observation in radian.
        els : 1d array
            The elevation array for the observation in radian.

        Returns
        ----------
        ra : 1d array
            Right ascension in radian.
        dec : 1d array
            Declination in radian.
        pa : 1d array
            Parallactic angle in radian.
        """
        ra, dec, pa = self.quaternion.offset_radecpa_applyquat(
            self.q, -azd, -eld)
        return ra, dec, pa

class Azel2Radec(object):
    """ Class to handle az/el <-> ra/dec conversion """
    def __init__(self, mjd, ut1utc,
                 lon=-67.786, lat=-22.958, height=5200.,
                 pressure=533.29, temp=273.15, humidity=0.1, epequi=2000.0):
        """
        This class is mostly a wrapper around slalib.
        The default parameters correspond to the Polarbear observation site.

        Parameters
        ----------
        mjd : float
            Date in MJD.
        ut1utc : float
            Time correction (UT1 - UTC) for the date.
        lon : float, optional
            Longitute of the site of observation in degree.
        lat : float, optional
            Latitude of the site of observation in degree.
        height : float, optional
            Altitude of the site of observation.
        pressure : float, optional
            Local atmospheric pressure at the site of observation in millibar.
        temp : float, optional
            Local ambient temperature at the site of observation in K.
        humidity : float, optional
            Local relative humidity at the site of observation
            (value between 0 and 1).
        epequi : float, optional
            Epoch of mean equinox to be used (Julian).

        """
        self.lon = lon * d2r
        self.lat = lat * d2r
        self.height = height
        self.pressure = pressure
        self.temp = temp
        self.humidity = humidity
        self.mjd = mjd

        self.epequi = epequi
        self.ut1utc = ut1utc

        self.updateaoprms(mjd)

    def updateaoprms(self, mjd, lapserate=0.0065, xpm=0.0, ypm=0.0):
        """
        Pre-compute the set of apparent to observed place parameters.
        See http://star-www.rl.ac.uk/docs/sun67.htx/sun67ss8.html

        Parameters
        ----------
        lapserate : float
            ta mere
        xpm : float
            polar motion coordinate x (radians)
        ypm : float
            polar motion coordinate y (radians)

        """
        wavelength = (299792458.0 / 150.0e9) * 1e6
        self.aoprms = slalib.sla_aoppa(mjd, self.ut1utc, self.lon, self.lat,
                                       self.height, xpm, ypm, self.temp,
                                       self.pressure, self.humidity,
                                       wavelength, lapserate)

    def azel2radecpa(self, mjd, az, el):
        """
        Given Az/El and time returns RA/Dec and parallactic angle.
        This routine does not return a precisely correct parallactic angle.

        Parameters
        ----------
        mjd : float
            Date in MJD.
        az : float
            Azimuth in radian.
        el : float
            Elevation in radian.

        Returns
        ----------
        ra : float
            Right ascension in radian.
        dec : float
            Declination in radian.
        pa : float
            Parallactic angle in radian.
        """
        zd = np.pi / 2 - el
        amprms = slalib.sla_mappa(self.epequi, mjd)
        self.aoprms = slalib.sla_aoppat(mjd, self.aoprms)

        ra_app1, dec_app1 = slalib.sla_oapqk('a', az, zd + 1e-8, self.aoprms)
        ra1, dec1 = slalib.sla_ampqk(ra_app1, dec_app1, amprms)
        ra_app2, dec_app2 = slalib.sla_oapqk('a', az, zd - 1e-8, self.aoprms)
        ra2, dec2 = slalib.sla_ampqk(ra_app2, dec_app2, amprms)
        pa = slalib.sla_dbear(ra1, dec1, ra2, dec2)
        ra = 0.5 * (ra1 + ra2)
        dec = 0.5 * (dec1 + dec2)

        return ra, dec, pa

    def radec2azel(self, mjd, ra, dec):
        """
        Given RA/Dec and time returns Az/El.

        Parameters
        ----------
        mjd : float
            Date in MJD.
        ra : float
            Right ascension in radian.
        dec : float
            Declination in radian.

        Returns
        ----------
        az : float
            Azimuth in radian.
        el : float
            Elevation in radian.
        """
        amprms = slalib.sla_mappa(self.epequi, mjd)
        self.aoprms = slalib.sla_aoppat(mjd, self.aoprms)
        ra_app, dec_app = slalib.sla_mapqkz(ra, dec, amprms)
        az, zd, a, b, c = slalib.sla_aopqk(ra_app, dec_app, self.aoprms)
        el = np.pi / 2 - zd
        return az, el

class Quaternion():
    """ Class to handle quaternions """
    def __init__(self, ra, dec, pa, v_ra_src, v_dec_src):
        """
        Once you have RA/Dec coordinates of the reference detector, one
        can efficiently use quaternions to compute each detector pointing in
        the focal plane. This is rather faster than the traditional method,
        but a bit more abstract and less intuitive unfortunately...

        Parameters
        ----------
        ra : 1d array
            Right ascension in radian.
        dec : 1d array
            Declination in radian.
        pa : 1d array
            Parallactic angle in radian.
        v_ra_src : float
            RA of the source (center of the patch for example).
        v_dec_src : float
            Dec of the source (center of the patch for example).

        """
        self.ra = ra
        self.dec = dec
        self.pa = pa
        self.v_ra_src = v_ra_src
        self.v_dec_src = v_dec_src

    def offset_radecpa_makequat(self):
        """
        Construct the quaternions from initial RA/Dec coordinates.

        Returns
        ----------
        q : array
            Quaternions array.
        """
        qra = euler_quatz(self.ra)

        qdec = euler_quaty(-self.dec)
        qpa = euler_quatx(-self.pa)

        qracen = euler_quatz(-self.v_ra_src)
        qdeccen = euler_quaty(self.v_dec_src)

        q = mult(qdec, qpa)
        q = mult(qra, q)
        q = mult(qracen, q)
        q = mult(qdeccen, q)

        return q

    def offset_radecpa_applyquat(self, q, azd, eld):
        """
        Apply pre-computed quaternions to obtain
        desired RA/Dec and parallactic angle from az/el of the detector.

        Parameters
        ----------
        q : array
            Quaternions array.
        azd : 1d array
            Azimuth of the detector.
        eld : 1d array
            Elevation of the detector.

        Returns
        ----------
        psi : 1d array
            Can be identified to RA (if input -azd, -eld)
        theta : 1d array
            Can be identified to -Dec (if input -azd, -eld). Note that -theta
            is returned (so Dec directly).
        phi : 1d array
            Can be identified to -PA (if input -azd, -eld). Note that -phi
            is returned (so PA directly).
        """
        assert len(q.shape) == 2, AssertionError("Wrong quaternion size!")
        assert q.shape[1] == 4, AssertionError("Wrong quaternion size!")
        qazd = euler_quatz(-azd)
        qeld = euler_quaty(-eld)

        qpix = mult(qazd, qeld)[0]

        # Inlining this is a 30x speed up
        seq = mult_fortran(q, qpix)

        assert seq.shape[1] == 4, AssertionError("Wrong size!")

        n = seq.shape[0]
        phi, theta, psi = quat_to_radecpa_fortran(seq)

        return psi, -theta, -phi

def radec2thetaphi(ra, dec):
    """
    Correspondance between RA/Dec and theta/phi coordinate systems.

    Parameters
    ----------
    ra : float or 1d array
        RA angle in radian.
    dec : float or 1d array
        Dec angle in radian.

    Returns
    ----------
    theta : float or 1d array
        Theta angle in radian
    phi : float or 1d array
        Phi angle in radian

    Examples
    ----------
    >>> theta, phi = radec2thetaphi(-np.pi/2., 0.)
    >>> print(round(theta, 2), round(phi, 2))
    1.57 -1.57
    """
    theta = np.pi / 2 - dec
    phi = ra
    return theta, phi

def mult(p, q):
    """
    Multiply arrays of quaternions,
    see: http://en.wikipedia.org/wiki/\
        Quaternions#Quaternions_and_the_geometry_of_R3

    Parameters
    ----------
    p : ndarray
        Array of quaternions of size (np, 4)
    q : ndarray
        Array of quaternions of size (nq, 4)

    Returns
    ----------
    pq : ndarray
        Array of size (np, 4)

    Examples
    ----------
    >>> mult(np.array([[3., 4., 5., 2.],
    ...     [2., 2., 2., 2.]]), np.array([1., 2., 3., 4.]))
    ... #doctest +NORMALIZE_WHITESPACE
    array([[ 16.,  16.,  28., -18.],
           [ 12.,   8.,  16.,  -4.]])
    """
    if p.ndim == 1 and q.ndim > 1:
        p = np.tile(p, (q.shape[0], 1))
    if q.ndim == 1 and p.ndim > 1:
        q = np.tile(q, (p.shape[0], 1))
    if q.ndim == 1 and p.ndim == 1:
        p = p.reshape((1, 4))
        q = q.reshape((1, 4))

    ps = p[:, 3]
    qs = q[:, 3]
    pv = p[:, :3]
    qv = q[:, :3]

    pq = np.empty_like(p)
    pq[:, 3] = ps * qs
    pq[:, 3] -= arraylist_dot(pv, qv).flatten()

    pq[:, :3] = ps[:, np.newaxis] * qv + \
        pv * qs[:, np.newaxis] + np.cross(pv, qv)

    return pq

def mult_fortran(p, q):
    """
    Inline version for when p is an array of quaternions
    and q is a single quaternion. Big speed-up.

    Parameters
    ----------
    p : ndarray
        Array of quaternions of size (np, 4)
    q : ndarray
        Array of quaternions of size (1, 4)

    Returns
    ----------
    pq : ndarray
        Array of size (np, 4)

    Examples
    ----------
    >>> mult_fortran(np.array([[3., 4., 5., 2.],
    ...     [2., 2., 2., 2.]]), np.array([1., 2., 3., 4.]))
    ... #doctest +NORMALIZE_WHITESPACE
    array([[ 16.,  16.,  28., -18.],
           [ 12.,   8.,  16.,  -4.]])
    """

    assert p.ndim == 2, AssertionError("Wrong size!")
    assert p.shape[1] == 4, AssertionError("Wrong size!")
    assert q.ndim == 1, AssertionError("Wrong size!")
    assert q.size == 4, AssertionError("Wrong size!")
    pq = np.zeros_like(p)
    n = p.shape[0]

    shape = pq.shape
    pq = pq.flatten()
    p = p.flatten()
    q = q.flatten()

    detector_pointing_f.mult_fortran_f(p, q, pq, n)

    pq = pq.reshape(shape)
    return pq

def arraylist_dot(a, b):
    """
    Dot product of ndarrays.
    p and q can have any size.
    Returns a column array.

    Parameters
    ----------
    a : array of arrays
        first list of arrays.
    b : array of arrays
        second list of arrays.

    Examples
    ----------
    >>> a = np.array([[1, 2, 3]])
    >>> b = np.array([[4, 6, 8], [3, 7, 1]])
    >>> arraylist_dot(a, b) #doctest: +NORMALIZE_WHITESPACE
    array([[40],
           [20]])
    """
    if a.ndim == 1 and b.ndim == 1:
        return np.array([[np.dot(a, b)]])
    else:
        return np.sum(a * b, axis=1)[:, np.newaxis]

def euler_quatx(alpha):
    """
    Generate quaternion units along x axis

    Parameters
    ----------
    alpha : float
        Polar angle in radian.

    Examples
    ----------
    >>> euler_quatx(alpha=np.pi/2.)
    array([ 0.70710678,  0.        ,  0.        ,  0.70710678])

    """
    alpha = np.asarray(alpha)
    z = np.zeros_like(alpha)
    c = np.cos(alpha * 0.5)
    s = np.sin(alpha * 0.5)
    return np.array([s, z, z, c]).T

def euler_quaty(alpha):
    """
    Generate quaternion units along y axis

    Parameters
    ----------
    alpha : float
        Polar angle in radian.

    Examples
    ----------
    >>> euler_quaty(alpha=np.pi/2.)
    array([ 0.        ,  0.70710678,  0.        ,  0.70710678])

    """
    alpha = np.asarray(alpha)
    z = np.zeros_like(alpha)
    c = np.cos(alpha * 0.5)
    s = np.sin(alpha * 0.5)
    return np.array([z, s, z, c]).T

def euler_quatz(alpha):
    """
    Generate quaternion units along z axis

    Parameters
    ----------
    alpha : float
        Polar angle in radian.

    Examples
    ----------
    >>> euler_quatz(alpha=np.pi/2.)
    array([ 0.        ,  0.        ,  0.70710678,  0.70710678])

    """
    alpha = np.asarray(alpha)
    z = np.zeros_like(alpha)
    c = np.cos(alpha * 0.5)
    s = np.sin(alpha * 0.5)
    return np.array([z, z, s, c]).T

def quat_to_radecpa_fortran(seq):
    """
    Routine to compute phi/theta/psi from a sequence
    of quaternions. Computation is done in fortran.

    WARNING: you still need to convert phi/theta/psi to get to RA/Dec/PA.

    Parameters
    ----------
    seq : array of arrays
        Array of quaternions.

    Returns
    ----------
    phi : 1d array
    theta : 1d array
    psi : 1d array
    """
    q1, q2, q3, q0 = seq.T
    n = q0.size
    phi = np.zeros_like(q0)
    theta = np.zeros_like(q0)
    psi = np.zeros_like(q0)

    detector_pointing_f.quat_to_radecpa_fortran_f(
        q0, q1, q2, q3, phi, theta, psi, n)
    return phi, theta, psi

def quat_to_radecpa_python(seq):
    """
    Routine to compute phi/theta/psi from a sequence
    of quaternions. Computation is done in Python.

    WARNING: you still need to convert phi/theta/psi to get to RA/Dec/PA.

    Parameters
    ----------
    seq : array of arrays
        Array of quaternions.

    Returns
    ----------
    phi : 1d array
    theta : 1d array
    psi : 1d array

    """
    q1, q2, q3, q0 = seq.T
    phi = np.arctan2(2 * (q0 * q1 + q2 * q3),
                     1. - 2. * (q1 * q1 + q2 * q2))
    theta = np.arcsin(2 * (q0 * q2 - q3 * q1))
    psi = np.arctan2(2 * (q0 * q3 + q1 * q2),
                     1. - 2. * (q2 * q2 + q3 * q3))
    return phi, theta, psi

def load_fake_pointing():
    """
    Load fake pointing parameters for testing purposes.

    Returns
    ----------
    allowed_params : list of string
        Name of the pointing model parameters used in `value_params`.
    value_params : 1d array
        Value of the pointing model parameters (see instrument.py).
        In degrees.
    az_enc : 1d array
        Encoder azimuth in radians.
    el_enc : 1d array
        Encoder elevation in radians.
    time : 1d array
        Encoder time (UTC) in mjd

    """
    allowed_params = 'ia ie ca an aw'
    value_params = [10.28473073, 8.73953334, -15.59771781,
                    -0.50977716, 0.10858016]
    az_enc = np.array([np.sin(2 * np.pi * i / 100)
                       for i in range(100)])
    el_enc = np.ones(100) * 0.5
    time = np.array([56293 + t/84000. for t in range(100)])

    return allowed_params, value_params, az_enc, el_enc, time


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/python
"""
Script to simulate the scan of a CMB experiment.

Author: Julien Peloton, j.peloton@sussex.ac.uk
"""
from __future__ import division, absolute_import, print_function

import os
import ephem
import numpy as np
import healpy as hp

from s4cmb.scanning_strategy_f import scanning_strategy_f

from pyslalib import slalib

## numerical constants
radToDeg = 180. / np.pi
sidDayToSec = 86164.0905

class ScanningStrategy():
    """ Class to handle the scanning strategy of the telescope """
    def __init__(self, nces=12, start_date='2013/1/1 00:00:00',
                 telescope_longitude='-67:46.816',
                 telescope_latitude='-22:56.396', telescope_elevation=5200.,
                 name_strategy='deep_patch', sampling_freq=30., sky_speed=0.4,
                 ut1utc_fn='s4cmb/data/ut1utc.ephem', language='python',
                 verbose=False):
        """
        A scanning strategy consists in defining the site of observation
        on earth for which we will make the observation, the region
        of the sky to observe, and the schedule of observations.

        Parameters
        ----------
        nces : int, optional
            Number of scans to generate.
        start_date : string, optional
            Starting date for observations. The format is: YYYY/M/D HH:MM:SS.
        telescope_longitude : string, optional
            Longitute (angle) of the telescope. String form: 0:00:00.0.
        telescope_latitude : string, optional
            Latitude (angle) of the telescope. String form: 0:00:00.0.
        telescope_elevation : float, optional
            Height above sea level (in meter).
        name_strategy : string, optional
            Name of a pre-defined scanning strategy to define the boundaries
            of the scan: elevation, azimuth, and time. Only available for the:
            moment name_strategy = deep_patch.
        sampling_freq : float, optional
            Sampling frequency of the bolometers in Hz.
        sky_speed : float, optional
            Azimuth speed of the telescope in deg/s.
        ut1utc_fn : string, optional
            File containing time correction to UTC (leap second).
            See http://tycho.usno.navy.mil/leapsec.html.
            This is not used here, but pass to the pointing module later on.
        language : string, optional
            Language used for core computations. For big experiments, the
            computational time can be big, and some part of the code can be
            speeded up by interfacing python with Fortran.
            Default is python (i.e. no interfacing and can be slow).
            Choose language=fortran otherwise. Note that for fortran codes you
            need first to compile it. See the setup.py or
            the provided Makefile.
        verbose : bool
            If True, print out several messages to ease the debug.
            Default is False.

        """
        self.nces = nces
        self.start_date = start_date
        self.name_strategy = name_strategy
        self.sampling_freq = sampling_freq
        self.sky_speed = sky_speed
        self.language = language

        self.ut1utc_fn = ut1utc_fn
        if not os.path.isfile(self.ut1utc_fn):
            url = 'http://tycho.usno.navy.mil/leapsec.html'
            msg = 'The path {} does not point to a valid file! see ' + \
                's4cmb/data/ut1utc.ephem provided with the package. ' + \
                'For more information, see {}.'
            raise Exception(msg.format(self.ut1utc_fn, url))

        self.verbose = verbose

        self.telescope_location = self.define_telescope_location(
            telescope_longitude, telescope_latitude, telescope_elevation)

        self.define_boundary_of_scan()

    def define_telescope_location(self, telescope_longitude='-67:46.816',
                                  telescope_latitude='-22:56.396',
                                  telescope_elevation=5200.):
        """
        Routine to define the site of observation on earth for which
        positions are to be computed. The location of the Polarbear telescope
        is entered as default.

        Parameters
        ----------
        telescope_longitude : str
            Longitute (angle) of the telescope. String form: 0:00:00.0.
        telescope_latitude : str
            Latitude (angle) of the telescope. String form: 0:00:00.0.
        telescope_elevation : float
            Height above sea level (in meter).

        Returns
        ----------
        location : Observer instance
            An `Observer` instance allows you to compute the positions of
            celestial bodies as seen from a particular latitude and longitude
            on the Earth's surface.

        Examples
        ----------
        >>> scan = ScanningStrategy()
        >>> telescope_location = scan.define_telescope_location()
        >>> telescope_location.elevation
        5200.0

        """
        location = ephem.Observer()
        location.long = telescope_longitude
        location.lat = telescope_latitude
        location.elevation = telescope_elevation

        return location

    def define_boundary_of_scan(self):
        """
        Given a pre-defined scanning strategy,
        define the boundaries of the scan: elevation, azimuth, and time.
        For a custom usage (advanced users), modify this routine.

        Examples
        ----------
        >>> scan = ScanningStrategy(name_strategy='deep_patch')
        >>> scan.elevation # doctest: +NORMALIZE_WHITESPACE
        [30.0, 45.5226, 47.7448, 49.967,
         52.1892, 54.4114, 56.6336, 58.8558,
         61.078, 63.3002, 65.5226, 35.2126]

        Only few scanning strategies are currently available:
        >>> scan = ScanningStrategy(name_strategy='shallow_patch')
        >>> scan = ScanningStrategy(name_strategy='small_aperture')
        >>> scan = ScanningStrategy(name_strategy='custom')
        >>> scan = ScanningStrategy(name_strategy='toto')
        ... # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        Traceback (most recent call last):
         ...
        ValueError: Only name_strategy = deep_patch or shallow_patch or
        custom are currently available. For another usage (advanced users),
        modify this routine.

        >>> scan.allowed_scanning_strategies
        ['deep_patch', 'shallow_patch', 'small_aperture', 'custom']
        """
        self.allowed_scanning_strategies = [
            'deep_patch',
            'shallow_patch',
            'small_aperture',
            'custom']

        if self.name_strategy == 'deep_patch':
            self.elevation = [30.0, 45.5226, 47.7448, 49.967,
                              52.1892, 54.4114, 56.6336, 58.8558,
                              61.078, 63.3002, 65.5226, 35.2126]

            self.az_min = [134.2263, 162.3532, 162.3532, 162.3532,
                           162.3532, 162.3532, 162.3532, 162.3532,
                           162.3532, 162.3532, 162.3532, 204.7929]

            self.az_max = [154.2263, 197.3532, 197.3532, 197.3532,
                           197.3532, 197.3532, 197.3532, 197.3532,
                           197.3532, 197.3532, 197.3532, 224.7929]

            self.begin_LST = ['17:07:54.84', '22:00:21.76', '22:00:21.76',
                              '22:00:21.76', '22:00:21.76', '22:00:21.76',
                              '22:00:21.76', '22:00:21.76', '22:00:21.76',
                              '22:00:21.76', '22:00:21.76', '2:01:01.19']

            self.end_LST = ['22:00:21.76', '02:01:01.19', '02:01:01.19',
                            '02:01:01.19', '02:01:01.19', '02:01:01.19',
                            '02:01:01.19', '02:01:01.19', '02:01:01.19',
                            '02:01:01.19', '02:01:01.19', '6:53:29.11']

            self.dec_min = None
            self.dec_max = None
            self.begin_RA = None
            self.end_RA = None
            self.orientation = None

            ## Center of the patch in RA/Dec
            self.ra_mid = 0.
            self.dec_mid = -57.5
        elif self.name_strategy == 'small_aperture':
            self.elevation = [50.] * 6

            self.az_min = [37.1, 282.8, 110.3,
                           197.0, 110.3, 197.0]

            self.az_max = [77.7, 322.9, 163.0,
                           250.2, 163.0, 250.2]

            self.begin_LST = ['08:30:00.00', '12:30:00.00', '16:30:00.00',
                              '02:30:00.00', '16:30:00.00', '22:30:00.00']

            self.end_LST = ['12:29:59.32', '16:29:59.32', '02:29:59.30',
                            '08:29:59.98', '22:29:59.99', '08:29:59.30']

            self.dec_min = None
            self.dec_max = None
            self.begin_RA = None
            self.end_RA = None
            self.orientation = None

            ## Center of the patch in RA/Dec
            self.ra_mid = 0.
            self.dec_mid = 0.
        elif self.name_strategy == 'shallow_patch':
            self.elevation = [30., 30., 30., 30., 30., 30.]

            self.dec_min = ["-5:00:00", "-5:00:00", "-15:00:00",
                            "-15:00:00", "-60:00:00", "-60:00:00"]

            self.dec_max = ["20:00:00", "20:00:00", "20:00:00",
                            "20:00:00", "-15:00:00", "-15:00:00"]

            self.begin_RA = ["7:40:00", "7:40:00", "20:00:00",
                             "20:00:00", "20:00:00", "20:00:00"]

            self.end_RA = ["15:20:00", "15:20:00", "5:40:00",
                           "5:40:00", "5:40:00", "5:40:00"]

            self.orientation = ['west', 'east', 'west', 'east', 'west', 'east']

            self.az_min = None
            self.az_max = None
            self.begin_LST = None
            self.end_LST = None

            ## Center of the patch in RA/Dec
            self.ra_mid = 0.
            self.dec_mid = 0.
        elif self.name_strategy == 'custom':
            self.elevation = None
            self.dec_min = None
            self.dec_max = None
            self.begin_RA = None
            self.end_RA = None
            self.orientation = None
            self.az_min = None
            self.az_max = None
            self.begin_LST = None
            self.end_LST = None

            ## Center of the patch in RA/Dec
            self.ra_mid = 0.
            self.dec_mid = 0.
        else:
            raise ValueError("Only name_strategy = deep_patch or " +
                             "shallow_patch or custom are " +
                             "currently available. For another usage " +
                             "(advanced users), modify this routine.")

    def run_one_scan(self, scan_file, scan_number):
        """
        Generate one observation (i.e. one CES) of the telescope.

        Parameters
        ----------
        scan_file : dictionary
            Empty dictionary which will contain the outputs of the scan.
        scan_number : int
            Index of the scan (between 0 and nces - 1).

        Returns
        ----------
        bool : bool
            Returns True if the scan has been generated, and False if the scan
            already exists on the disk.
        """
        ## Check if we have too much/enough information to make a scan
        msg = "You cannot specify azimuth and declination!"
        assert (getattr(self, 'az_min') and not getattr(self, 'dec_min')) or \
            (not getattr(self, 'az_min') and getattr(self, 'dec_min')), msg
        assert (getattr(self, 'az_max') and not getattr(self, 'dec_max')) or \
            (not getattr(self, 'az_max') and getattr(self, 'dec_max')), msg

        msg = "You cannot specify LST and RA!"
        assert (getattr(self, 'begin_LST') and not getattr(self, 'begin_RA')) or \
            (not getattr(self, 'begin_LST') and getattr(self, 'begin_RA')), msg
        assert (getattr(self, 'end_LST') and not getattr(self, 'end_RA')) or \
            (not getattr(self, 'end_LST') and getattr(self, 'end_RA')), msg

        if getattr(self, 'az_min') and getattr(self, 'az_max'):
            msg = 'You need to define timing bounds!'
            assert getattr(self, 'begin_LST') is not None, msg
            assert getattr(self, 'end_LST') is not None, msg
            azLST = True
            RADEC = False
            if self.verbose:
                print("Using azimuth and LST bounds")

        elif getattr(self, 'dec_min') and getattr(self, 'dec_max'):
            msg = 'You need to define RA bounds!'
            assert getattr(self, 'begin_RA') is not None, msg
            assert getattr(self, 'end_RA') is not None, msg
            msg = 'You need to define orientation of scan (east/west)!'
            assert getattr(self, 'orientation') is not None, msg
            azLST = False
            RADEC = True
            if self.verbose:
                print("Using RA and Dec bounds")

        ## Figure out the elevation to run the scan
        el = self.elevation[scan_number]

        ## Define the sampling rate in Hz
        sampling_freq = self.sampling_freq

        #########################################################
        ## Define geometry of the scan
        #########################################################
        if azLST:
            ## Define geometry of the scan by figuring out the azimuth bounds
            az_mean = (
                self.az_min[scan_number] + self.az_max[scan_number]) * 0.5
            az_throw = (self.az_max[scan_number] -
                        self.az_min[scan_number]) / np.cos(el / radToDeg)

        elif RADEC:
            ## If given bounds in declination, make bounds in azimuth
            ## note there is no sanity checking here!
            az_array = np.linspace(0., 180., endpoint=True, num=360.)
            ra_array = np.zeros(az_array.shape)
            dec_array = np.zeros(az_array.shape)
            dec_min = ephem.degrees(self.dec_min[scan_number])
            dec_max = ephem.degrees(self.dec_max[scan_number])
            if(self.orientation[scan_number] == 'west'):
                az_array += 180.

            for i in range(0, az_array.shape[0]):
                ra_array[i], dec_array[i] = \
                    self.telescope_location.radec_of(
                        az_array[i] / radToDeg, el / radToDeg)

            az_allowed = np.asarray(
                [az_array[i] for i in range(0, az_array.shape[0])
                    if (dec_array[i] > dec_min and dec_array[i] < dec_max)])

            if (az_allowed.shape[0] < 2):
                ms = 'Invalid combination of declination bounds and elevation.'
                print(ms)
                exit(1)

            az_max = np.max(az_allowed)
            az_min = np.min(az_allowed)
            az_mean = (az_min + az_max) * 0.5
            az_throw = (az_max - az_min)

        #########################################################
        ## Define the timing bounds!
        #########################################################
        if azLST:
            LST_now = float(
                self.telescope_location.sidereal_time()) / (2 * np.pi)
            begin_LST = float(
                ephem.hours(self.begin_LST[scan_number])) / (2 * np.pi)
            end_LST = float(
                ephem.hours(self.end_LST[scan_number])) / (2 * np.pi)

            if (begin_LST > end_LST):
                begin_LST -= 1.

            ## Reset the date to correspond to the sidereal time to start
            self.telescope_location.date -= (
                (LST_now - begin_LST) * sidDayToSec) * ephem.second

            ## Figure out how long to run the scan for
            num_pts = int((end_LST - begin_LST) * sidDayToSec * sampling_freq)

        if RADEC:
            self.telescope_location.horizon = el * ephem.degree

            # Define a fixed source at the ra, dec that we want to scan
            target_min_ra = ephem.FixedBody()
            target_max_ra = ephem.FixedBody()

            # Figure out where we are looking now
            ra_target, dec_target = self.telescope_location.radec_of(
                az_mean / radToDeg, el / radToDeg)

            # Instantiate the targets
            target_min_ra._dec = dec_target
            target_max_ra._dec = dec_target
            target_min_ra._ra = self.begin_RA[scan_number]
            target_max_ra._ra = self.end_RA[scan_number]

            # Compute initial RA
            target_min_ra.compute(self.telescope_location)
            target_max_ra.compute(self.telescope_location)
            if(self.orientation[scan_number] == 'east'):
                self.telescope_location.date =  \
                    self.telescope_location.next_rising(target_min_ra)

                # Recompute coodinates in the light of change of date
                target_min_ra.compute(self.telescope_location)
                target_max_ra.compute(self.telescope_location)

                ## Update number of time samples for the scan
                num_pts = int(
                    (self.telescope_location.next_rising(
                        target_max_ra) - self.telescope_location.date) /
                    ephem.second * sampling_freq)

            if(self.orientation[scan_number] == 'west'):
                self.telescope_location.date =  \
                    self.telescope_location.next_setting(target_min_ra)

                # Recompute coodinates in the light of change of date
                target_min_ra.compute(self.telescope_location)
                target_max_ra.compute(self.telescope_location)

                ## Update number of time samples for the scan
                num_pts = int(
                    (self.telescope_location.next_setting(
                        target_max_ra) - self.telescope_location.date) /
                    ephem.second * sampling_freq)

        ## Run the scan!
        pb_az_dir = 1.
        upper_az = az_mean + az_throw / 2.
        lower_az = az_mean - az_throw / 2.
        az_speed = self.sky_speed / np.cos(el / radToDeg)
        running_az = az_mean

        ## Initialize arrays
        pb_az_array = np.zeros(num_pts)
        pb_mjd_array = np.zeros(num_pts)
        pb_ra_array = np.zeros(num_pts)
        pb_dec_array = np.zeros(num_pts)
        pb_el_array = np.ones(num_pts) * el

        ## Loop over time samples
        # begin_lst = str(self.telescope_location.sidereal_time())
        # Pad scans 10 seconds on either side
        time_padding = 10.0 / 86400.0

        ## Start of the scan
        pb_az_array[0] = running_az
        pb_mjd_array[0] = date_to_mjd(self.telescope_location.date)

        ## Initialize the time
        scan_file['firstmjd'] = pb_mjd_array[0] - time_padding

        ## Update before starting the loop
        running_az += az_speed * pb_az_dir / sampling_freq
        self.telescope_location.date += ephem.second / sampling_freq

        if self.language == 'python':
            for t in range(1, num_pts):
                ## Set the Azimuth and time
                pb_az_array[t] = running_az

                pb_ra_array[t], pb_dec_array[t] = \
                    self.telescope_location.radec_of(
                    pb_az_array[t] * np.pi / 180.,
                    pb_el_array[t] * np.pi / 180.)

                ## Case to change the direction of the scan
                if(running_az > upper_az):
                    pb_az_dir = -1.
                elif(running_az < lower_az):
                    pb_az_dir = 1.

                running_az += az_speed * pb_az_dir / sampling_freq

                ## Increment the time by one second / sampling rate
                pb_mjd_array[t] = pb_mjd_array[t-1] + \
                    ephem.second / sampling_freq

                ## Increment the time by one second / sampling rate
                self.telescope_location.date += ephem.second / sampling_freq

        elif self.language == 'fortran':
            second = 1./24./3600.
            scanning_strategy_f.run_one_scan_f(
                pb_az_array, pb_mjd_array,
                running_az, upper_az, lower_az, az_speed, pb_az_dir,
                second, sampling_freq, num_pts)

        ## Do not use that for precision - it truncates values
        self.telescope_location.date += num_pts * ephem.second / sampling_freq

        ## Save in file
        scan_file['nces'] = self.nces
        scan_file['CES'] = scan_number
        scan_file['sample_rate'] = sampling_freq
        scan_file['sky_speed'] = self.sky_speed
        scan_file['lastmjd'] = pb_mjd_array[-1] + time_padding

        scan_file['azimuth'] = pb_az_array * np.pi / 180
        scan_file['elevation'] = pb_el_array * np.pi / 180
        scan_file['clock-utc'] = pb_mjd_array

        scan_file['RA'] = pb_ra_array
        scan_file['Dec'] = pb_dec_array

        scan_file['nts'] = len(pb_mjd_array)

        if self.verbose:
            print('+-----------------------------------+')
            print(' CES starts at %s and finishes at %s' % (
                mjd_to_greg(scan_file['firstmjd']),
                mjd_to_greg(scan_file['lastmjd'])))
            print(' It lasts %.3f hours' % (
                (scan_file['lastmjd'] - scan_file['firstmjd']) * 24))
            print('+-----------------------------------+')

        ## Add one day before the next CES (to avoid conflict of time)
        self.telescope_location.date += 24 * ephem.second * 3600

        ## Add the scan into the instance
        # self._update('scan{}'.format(scan_number), scan_file)

        return True

    def run(self):
        """
        Generate all the observations (i.e. all CES) of the telescope.

        Examples
        ----------
        >>> scan = ScanningStrategy(sampling_freq=1., nces=2,
        ...     name_strategy='deep_patch')
        >>> scan.run()
        >>> print(scan.scan0['firstmjd'], scan.scan0['lastmjd'])
        56293.6202546 56293.8230093

        By default, the language used for the core computation is the Python.
        It can be quite slow for heavy configuration, and one can set up
        the language to fortran for speeding up the computation (x1000).
        Note that for using fortran codes you need first to compile it.
        See the setup.py or the provided Makefile.
        >>> scan = ScanningStrategy(sampling_freq=1., nces=2,
        ...     language='fortran', name_strategy='shallow_patch')
        >>> scan.run()
        >>> print(round(scan.scan0['firstmjd'], 2))
        56293.37

        Note that you can create your own scanning strategy. First choose
        the custom ones (set everything to None):
        >>> scan = ScanningStrategy(sampling_freq=1., nces=2,
        ...     language='fortran', name_strategy='custom')

        And then define your own parameters. Example:
        >>> scan.nces = 1
        >>> scan.elevation = [30.]
        >>> scan.az_min = [60.]
        >>> scan.az_max = [100.]
        >>> scan.begin_LST = ['17:00:00.00']
        >>> scan.end_LST = ['22:00:00.00']
        >>> scan.run()

        Note that you To create a scanning strategy within s4cmb,
        you need to specify either
        * Elevations + minimum and maximum azimuths (spatial bounds) +
            begining and end Local Sidereal Times (timing bounds)
        * Elevations + minimum and maximum declinations +
            begining and end Right Ascensions (spatial & timing bounds) +
            orientations (east/west)

        """
        ## Initialise the date and loop over CESes
        self.telescope_location.date = self.start_date
        for CES_position in range(self.nces):
            ## Initialise the starting date of observation
            ## It will be updated then automatically
            setattr(self, 'scan{}'.format(CES_position), {})

            # Create the scan strategy
            self.run_one_scan(
                getattr(self, 'scan{}'.format(CES_position)), CES_position)

    def visualize_my_scan(self, nside, reso=6.9, xsize=900, rot=[0, -57.5],
                          nfid_bolometer=6000, fp_size=180., boost=1.,
                          fullsky=False):
        """
        Simple map-making: project time ordered data into sky maps for
        visualisation. It works only in pure python (i.e. if you set
        language='python' when initialising the scanning_strategy class).

        Parameters
        ----------
        nside : int
            Resolution of the healpix map.
        reso : float
            Resolution of the projected map (pixel size) in arcmin.
        xsize : int
            Number of pixel per row for the projected map.
        rot : 1d array of float
            Coordinates of the center of
            the projected map [lon, lat] in degree
        nfid_bolometer : int
            Fiducial number of bolometers for visualisation.
            By default, the scans are done for one reference bolometer.
        fp_size : float
            Size of the focal plane on the sky, in arcmin.
        boost : int
            boost factor to artificially increase the number of hits.
            It doesn't change the shape of the survey (just the amplitude).

        Outputs
        ----------
            * nhit_loc: 1D array, sky map with cumulative hit counts

        Examples
        ----------
        >>> import matplotlib
        >>> matplotlib.use("Agg") ## remove this if you want to display
        >>> scan = ScanningStrategy(sampling_freq=1., nces=1)
        >>> scan.run()
        >>> scan.visualize_my_scan(512)

        """
        import pylab as pl
        if self.language != 'python':
            raise ValueError("Visualisation is available only in pure " +
                             "python because we do not provide (yet) " +
                             "RA and Dec in fortran. Relaunch " +
                             "using language='python' in the class " +
                             "ScanningStrategy.")

        npix = hp.pixelfunc.nside2npix(nside)
        nhit = np.zeros(npix)
        for scan_number in range(self.nces):
            scan = getattr(self, 'scan{}'.format(scan_number))

            num_pts = len(scan['clock-utc'])
            pix_global = hp.pixelfunc.ang2pix(
                nside, (np.pi/2.) - scan['Dec'], scan['RA'])

            ## Boresight pointing healpix maps
            nhit_loc = np.zeros(npix, dtype=np.int32)
            scanning_strategy_f.mapmaking(
                pix_global, nhit_loc, npix, num_pts)

            ## Fake large focal plane with many bolometers for visualisation.
            nhit_loc = convolve_focalplane(nhit_loc, nfid_bolometer,
                                           fp_size, boost)

            nhit += nhit_loc

        if self.verbose:
            print('Stats: nhits = {}/{} (fsky={}%), max hit = {}'.format(
                len(nhit[nhit > 0]),
                len(nhit),
                round(len(nhit[nhit > 0])/len(nhit) * 100, 2),
                int(np.max(nhit))))

        nhit[nhit == 0] = hp.UNSEEN
        if not fullsky:
            hp.gnomview(nhit, rot=rot, reso=reso, xsize=xsize,
                        cmap=pl.cm.viridis,
                        title='nbolos = {}, '.format(nfid_bolometer) +
                        'fp size = {} arcmin, '.format(fp_size) +
                        'nhit boost = {}'.format(boost))
        else:
            hp.mollview(nhit, rot=rot, cmap=pl.cm.viridis,
                        title='nbolos = {}, '.format(nfid_bolometer) +
                        'fp size = {} arcmin, '.format(fp_size) +
                        'nhit boost = {}'.format(boost))
        hp.graticule(verbose=self.verbose)

        pl.show()

    def _update(self, name, value):
        """
        Wrapper around setattr function.
        Set a named attribute on an object.

        Parameters
        ----------
        name : string
            The name of the new attribute
        value : *
            The value of the attribute.
        """
        setattr(self, name, value)

def convolve_focalplane(bore_nhits, nbolos, fp_radius_amin, boost):
    """
    Given a hit count map,
    perform the focal plane convolution (that is generate the hit map as
    if there were `nbolos`, and boosted).
    Original author: Neil Goeckner-Wald.
    Modifications by Julien Peloton.

    Parameters
    ----------
    bore_nhits : 1D array
        number of hits per pixel for the reference detector.
    nbolos : int
        total number of bolometers desired.
    fp_radius_amin : float
        radius of the focal plane in arcmin.
    boost : float
        boost factor to artificially increase the number of hits.
        It doesn't change the shape of the survey (just the amplitude).

    Returns
    ----------
    focalplane_nhits : 1D array
        Number of hits for the all the detectors.

    Examples
    ----------
    >>> bore_nhits = np.zeros(hp.nside2npix(128))

    ## Put a patch in the center
    >>> bore_nhits[hp.query_disc(128, hp.ang2vec(np.pi/2, 0.),
    ...     radius=10*np.pi/180.)] = 1.

    ## Increase the number of detector (x100) and make a boost (x10)
    >>> conv_bore_nhits = convolve_focalplane(bore_nhits, nbolos=100,
    ...     fp_radius_amin=180, boost=10)
    >>> print(round(np.max(bore_nhits), 2), round(np.max(conv_bore_nhits), 2))
    1.0 1003.87
    """
    # Now we want to make the focalplane maps
    focalplane_nhits = np.zeros(bore_nhits.shape)

    # Resolution of our healpix map
    nside = hp.npix2nside(focalplane_nhits.shape[0])
    resol_amin = hp.nside2resol(nside, arcmin=True)
    fp_rad_bins = int(fp_radius_amin * 2. / resol_amin)
    fp_diam_bins = (fp_rad_bins * 2) + 1

    # Build the focal plane model and a list of offsets
    (x_fp, y_fp) = np.array(
        np.unravel_index(
            range(0, fp_diam_bins**2),
            (fp_diam_bins, fp_diam_bins))).reshape(
                2, fp_diam_bins, fp_diam_bins) - (fp_rad_bins)
    fp_map = ((x_fp**2 + y_fp**2) < (fp_rad_bins)**2)

    bolo_per_pix = nbolos / float(np.sum(fp_map))

    dRA = np.ndarray.flatten(
        (x_fp[fp_map].astype(float) * fp_radius_amin) / (
            fp_rad_bins * 60. * (180. / (np.pi))))
    dDec = np.ndarray.flatten(
        (y_fp[fp_map].astype(float) * fp_radius_amin) / (
            fp_rad_bins * 60. * (180. / (np.pi))))

    pixels_global = np.array(np.where(bore_nhits != 0)[0], dtype=int)
    for n in pixels_global:
        n = int(n)

        # Compute pointing offsets
        (theta_bore, phi_bore) = hp.pix2ang(nside, n)
        phi = phi_bore + dRA * np.sin(theta_bore)
        theta = theta_bore + dDec

        pixels = hp.ang2pix(nside, theta, phi)
        npix_loc = len(pixels)

        ## Necessary because the values in pixels aren't necessarily unique
        ## This is a poor design choice and should probably be fixed
        scanning_strategy_f.convolve_focalplane_f(
            bore_nhits[n], focalplane_nhits, pixels,
            bolo_per_pix, boost, npix_loc)

    return focalplane_nhits

## Here are a bunch of routines to handle dates...

def date_to_mjd(date):
    """
    Convert date in ephem.date format to MJD.

    Parameters
    ----------
    date : ephem.Date
        Floating point value used by ephem to represent a date.
        The value is the number of days since 1899 December 31 12:00 UT. When
        creating an instance you can pass in a Python datetime instance,
        timetuple, year-month-day triple, or a plain float.
        Run str() on this object to see the UTC date it represents.
        ...
        WTF?

    Returns
    ----------
    mjd : float
        Date in the format MJD.

    Examples
    ----------
    >>> e = ephem.Observer()
    >>> e.date = 0.0 ## 1899 December 31 12:00 UT
    >>> mjd = date_to_mjd(e.date)
    >>> print('DATE={} ->'.format(round(e.date, 2)),
    ...     'MJD={}'.format(round(mjd, 2)))
    DATE=0.0 -> MJD=15019.5
    """
    return greg_to_mjd(date_to_greg(date))

def date_to_greg(date):
    """
    Convert date in ephem.date format to gregorian date.

    Parameters
    ----------
    date : ephem.Date
        Floating point value used by ephem to represent a date.
        The value is the number of days since 1899 December 31 12:00 UT. When
        creating an instance you can pass in a Python datetime instance,
        timetuple, year-month-day triple, or a plain float.
        Run str() on this object to see the UTC date it represents.
        ...
        WTF?

    Returns
    ----------
    greg : string
        Gregorian date (format: YYYYMMDD_HHMMSS)

    Examples
    ----------
    >>> e = ephem.Observer()
    >>> e.date = 0.0 ## 1899 December 31 12:00 UT
    >>> greg = date_to_greg(e.date)
    >>> print('DATE={} ->'.format(round(e.date, 2)),
    ...     'GREG={}'.format(greg))
    DATE=0.0 -> GREG=18991231_120000
    """
    date_ = str(date)
    date_ = str(date.datetime())
    greg = date_.split('.')[0].replace('-',
                                       '').replace(':',
                                                   '').replace(' ',
                                                               '_')

    return greg

def greg_to_mjd(greg):
    """
    Convert gregorian date into MJD.

    Parameters
    ----------
    greg : string
        Gregorian date (format: YYYYMMDD_HHMMSS)

    Returns
    ----------
    mjd : float
        Date in the format MJD.

    Examples
    ----------
    >>> greg = '19881103_000000'
    >>> mjd = greg_to_mjd(greg)
    >>> print('GREG={} ->'.format(greg), 'MJD={}'.format(round(mjd, 2)))
    GREG=19881103_000000 -> MJD=47468.0
    """
    year = int(greg[:4])
    month = int(greg[4:6])
    day = int(greg[6:8])
    hour = int(greg[9:11])
    minute = int(greg[11:13])
    second = int(greg[13:15])

    fracday, status = slalib.sla_dtf2d(hour, minute, second)
    mjd, status = slalib.sla_cldj(year, month, day)
    mjd += fracday

    return mjd

def mjd_to_greg(mjd):
    """
    Convert MJD into gregorian date.

    Parameters
    ----------
    mjd : float
        Date in the format MJD.

    Returns
    ----------
    greg : string
        Gregorian date (format: YYYYMMDD_HHMMSS)

    Examples
    ----------
    >>> mjd = greg_to_mjd('19881103_000000')
    >>> greg = mjd_to_greg(mjd)
    >>> print('MJD={} ->'.format(round(mjd, 2)), 'GREG={}'.format(greg))
    MJD=47468.0 -> GREG=19881103_000000
    """
    year, month, day, fracday, baddate = slalib.sla_djcl(mjd)

    if baddate:
        raise ValueError(BadMJD)

    sign, (hour, minute, second, frac) = slalib.sla_dd2tf(2, fracday)

    s = '{:4d}{:2d}{:2d}_{:2d}{:2d}{:2d}'.format(
        year, month, day, hour, minute, second)
    s = s.replace(' ', '0')

    return s


if __name__ == "__main__":
    import doctest
    doctest.testmod()
